
rm(list = ls())

source("./prot_funct.R")
library(data.table)
library(parallel)

aa.set = c("G","A","S","T","C","V","L","I","M","P","F","Y","W","D","E","N","Q","H","K","R")
aa.set = aa.set[order(aa.set)]
onehot.mat = diag(20)
colnames(onehot.mat) = rownames(onehot.mat) = aa.set


###Load in parameters

args = commandArgs(T)

tcr.name = args[1]
rawdata.path = args[2]
peptide.length = as.numeric(args[3])
out.path = args[4]

##Example arguments for exploration / debugging
# tcr.name = "TCR1"
# rawdata.path = "~/Downloads/"
# peptide.length = 9
# out.path = "~/Downloads"


tcr.eval = fread(paste0(rawdata.path,"/",tcr.name,"_eval_table.csv"),data.table = F)
tcr.eval = tcr.eval[nchar(tcr.eval[,"Sequence"]) == peptide.length,]
tcr.eval = tcr.eval[order(tcr.eval[,"Round_3"],decreasing=T),]



pfm.bg = create.pfm.onehot(
  blosum.to.roundseq(as.character(tcr.eval[,1]),sub.mat = onehot.mat)
)

tcr.eval.enrich = tcr.eval[tcr.eval$Round_3 > 0,]

pfm = create.pfm.onehot(
  blosum.to.roundseq(as.character(tcr.eval.enrich[,1]),sub.mat = onehot.mat)
)
pwm = log10((pfm + .05) / (pfm.bg + .05))

write.table(pwm,paste0(out.path,"/",tcr.name,"_pwm.csv")
            ,row.names=F,col.names=F,quote=F,sep=",")

