
rm(list = ls())

# setwd("~/Dropbox/Publications/TCR/Code_copy_2/")
source("./prot_funct.R")
library(data.table)
library(parallel)

aa.set = c("G","A","S","T","C","V","L","I","M","P","F","Y","W","D","E","N","Q","H","K","R")
aa.set = aa.set[order(aa.set)]
onehot.mat = diag(20)
colnames(onehot.mat) = rownames(onehot.mat) = aa.set

# Set Parameters ----------------------------------------------------------

args = commandArgs(T)

tcr.name = args[1]
rawdata.path = args[2]
round.num = as.numeric(args[3])
peptide.length = as.numeric(args[4])
min.count = as.numeric(args[5])
out.path = args[6]

##Example arguments for exploration / debugging
# tcr.name = "test_tcr"
# rawdata.path = "./TCR1_Selection.csv"
# round.num = 4
# peptide.length = 9
# min.count = 3
# out.path = "./test_out2"

###Create the Eval table, which stores information about each peptide, but does not have
#the onehot encoding

tcr.eval = as.data.frame(fread(rawdata.path))
tcr.eval = tcr.eval[nchar(tcr.eval[,1]) == peptide.length,]
colnames(tcr.eval) = c("Sequence",paste0("Round_",0:round.num))
tcr.eval = as.data.frame(data.table(tcr.eval)[,lapply(.SD,sum),by = "Sequence"])
rownames(tcr.eval) = tcr.eval[,1]

tcr.eval = na.omit(tcr.eval)

###Normalize rounds so that there is an equal number of total counts per round.
norm.factor = mean(colSums(tcr.eval[,2:(round.num + 2)]))
tcr.eval[,2:(round.num + 2)] = apply(tcr.eval[,2:(round.num + 2)],2
                                                      ,function(x) x / sum(x) * norm.factor)
tcr.eval[,2:(round.num + 2)] = round(tcr.eval[,2:(round.num + 2)] )

tcr.eval = tcr.eval[rowSums(tcr.eval[,2:(round.num + 2)]) >= min.count,]



tcr.eval[,2:ncol(tcr.eval)] = log2(tcr.eval[,2:ncol(tcr.eval)] + 1)
tcr.eval = na.omit(tcr.eval)


###Because code looks for values in a specific column that assumes 4 rounds, add in dummy rounds 
if(round.num < 4){
  round.add = setdiff(1:4,1:round.num)
  for(i in round.add){
    tcr.eval[,paste0("Round_",i)] = 0
  }
}

##Compute fitness
tcr.eval$fitness = compute.trend(tcr.eval[,2:(2 + round.num)],cores = 6)
tcr.eval = tcr.eval[order(tcr.eval$fitness,decreasing=T),]

write.table(tcr.eval,paste0(out.path,"/",tcr.name,"_eval_table.csv"),sep=",",row.names=F)



round.seq.oh = cbind(blosum.to.roundseq(rownames(tcr.eval),sub.mat = onehot.mat),tcr.eval)


write.table(round.seq.oh[,-which(colnames(round.seq.oh) == "Sequence")]
            ,paste0(out.path,"/",tcr.name,"_onehot.csv")
            ,sep=",",row.names=F,quote=F,col.names=F)

