
source("./prot_funct.R")
library(data.table)
library(parallel)

aa.set = c("G","A","S","T","C","V","L","I","M","P","F","Y","W","D","E","N","Q","H","K","R")
aa.set = aa.set[order(aa.set)]
onehot.mat = diag(20)
colnames(onehot.mat) = rownames(onehot.mat) = aa.set


###Load in parameters


###Guide values -- use these to explore / debug code
# tcr.name = "TCR1"
# prot.name = "Human"
# pwm.path = "~/Downloads/"
# proteome.path = "~/Downloads"
# pred.num = 100
# prot.chunks = 15
# peptide.length = 9
# out.path = "~/Downloads"
# anchor.seq = "[A-Z][IMLV][A-Z][A-Z][A-Z][A-Z][A-Z][A-Z][IMLV]"

args = commandArgs(T)

tcr.name = args[1]
prot.name = args[2]
pwm.path = args[3]
proteome.path = args[4]
pred.num = as.numeric(args[5])
prot.chunks = as.numeric(args[6])
peptide.length = as.numeric(args[7])
out.path = args[8]
anchor.seq = args[9]


##Fixed parameters
feature.end = peptide.length * 20
response.col = "Round_3"

##load the PWM and convert it to a vector for matrix multiplication
pwm = as.numeric(as.matrix(read.csv(paste0(pwm.path,"/",tcr.name,"_pwm.csv")
                                    ,header=F)))




##Initialize a matrix to store the scores and the index values that map to the specific peptide 
#in the proteome
prot.preds = as.data.frame(matrix(0,prot.chunks * pred.num,2))
colnames(prot.preds) = c("index","Predicted.Score")

iter = 1
for(i in 1:prot.chunks){
  
  print(paste(i,iter))
  temp.chunk = as.data.frame(fread(
    paste0(proteome.path,"/",prot.name,"_proteome_onehot_",i,".csv")
    ,header=F
  ))
  
  preds.table = data.frame(
    index = temp.chunk[,ncol(temp.chunk)]
    ,preds = as.matrix(temp.chunk[,1:feature.end]) %*% pwm
  )
  
  preds.table = preds.table[order(preds.table$preds,decreasing = T),]
  prot.preds[iter:(iter + min(pred.num,nrow(preds.table))- 1),] = 
    preds.table[1:min(pred.num,nrow(preds.table)),]
  iter = iter + min(pred.num,nrow(preds.table))
}

prot.preds = prot.preds[prot.preds$index != 0,]


##Map scores back to the human proteome

load(paste0(proteome.path,"/",prot.name,"_kmerset.Rdata"))
kmer.table$kmer.seq = substring(kmer.table$kmer.seq,1,peptide.length)

indexed.preds = merge(prot.preds,kmer.table,by = "index")

indexed.preds$index = NULL
indexed.preds = indexed.preds[order(indexed.preds$Predicted.Score,decreasing= T),]
if(!is.null(anchor.seq)){ 
  indexed.preds$anchor.match = grepl(anchor.seq,indexed.preds$kmer.seq)
  indexed.preds = indexed.preds[order(indexed.preds$anchor.match,decreasing= T),]
}


write.table(indexed.preds,paste0(out.path,"/",tcr.name,"_indexed_PWM_predictions.csv")
            ,row.names=F,sep=",")
