source("./prot_funct.R")
library(data.table)
library(parallel)

aa.set = c("G","A","S","T","C","V","L","I","M","P","F","Y","W","D","E","N","Q","H","K","R")
aa.set = aa.set[order(aa.set)]
onehot.mat = diag(20)
colnames(onehot.mat) = rownames(onehot.mat) = aa.set


###Load in parameters


###Guide values -- use these to explore / debug code
# tcr.name = "TCR1"
# prot.name = "Human"
# rawdata.path = "~/Downloads/"
# proteome.path = "~/Downloads"
# pred.num = 100
# prot.chunks = 15
# peptide.length = 9
# out.path = "~/Downloads"
# anchor.seq = "[A-Z][IMLV][A-Z][A-Z][A-Z][A-Z][A-Z][A-Z][IMLV]"

args = commandArgs(T)

tcr.name = args[1]
prot.name = args[2]
rawdata.path = args[3]
proteome.path = args[4]
pred.num = as.numeric(args[5])
prot.chunks = as.numeric(args[6])
peptide.length = as.numeric(args[7])
out.path = args[8]
anchor.seq = args[9]



prot.preds = as.data.frame(matrix(0,prot.chunks * pred.num,2))
colnames(prot.preds) = c("index","Predicted.Score")

iter = 1
for(i in 1:prot.chunks){
  
  print(paste(i,iter))
  temp.chunk = fread(paste0(rawdata.path,"/",prot.name,"_",tcr.name,"_DL_scored_",i,".csv"),data.table = F)
  

  temp.chunk = temp.chunk[order(temp.chunk[,2],decreasing = T),]
  prot.preds[iter:(iter + min(pred.num,nrow(temp.chunk))- 1),] = 
    temp.chunk[1:min(pred.num,nrow(temp.chunk)),]
  iter = iter + min(pred.num,nrow(temp.chunk))
}


##Map scores back to the human proteome

load(paste0(proteome.path,"/",prot.name,"_kmerset.Rdata"))
kmer.table$kmer.seq = substring(kmer.table$kmer.seq,1,peptide.length)

indexed.preds = merge(prot.preds,kmer.table,by = "index")

indexed.preds$index = NULL
indexed.preds = indexed.preds[order(indexed.preds$Predicted.Score,decreasing= T),]
if(!is.null(anchor.seq)){ 
  indexed.preds$anchor.match = grepl(anchor.seq,indexed.preds$kmer.seq)
  indexed.preds = indexed.preds[order(indexed.preds$anchor.match,decreasing= T),]
}

write.table(indexed.preds,paste0(out.path,"/",tcr.name,"_indexed_DL_predictions.csv")
            ,row.names=F,sep=",")
