

##Functions and initialization
source("./prot_funct.R")
library(data.table)
library(parallel)


make.kmer = function(seq,kmer.length){
  
  sst = strsplit(seq,split="")[[1]]
  kmer.set = paste(sst[1:(nchar(seq) - kmer.length + 1)]
                   ,sst[
                     (1:(nchar(seq) - kmer.length + 1)) + 1]
                   ,sep="")
  
  
  for(i in 2:(kmer.length - 1)){
    # print(i)
    kmer.set = paste(kmer.set
                     ,sst[
                       (1:(nchar(seq) - kmer.length + 1)) + i]
                     ,sep="")
  }
  
  return(kmer.set)
  
}


aa.set = c("G","A","S","T","C","V","L","I","M","P","F","Y","W","D","E","N","Q","H","K","R")
aa.set = aa.set[order(aa.set)]
onehot.mat = diag(20)
colnames(onehot.mat) = rownames(onehot.mat) = aa.set



##Bring in command line arguments
args = commandArgs(T)

prot.name = args[1]
prot.path = args[2]
chunk.length = as.numeric(args[3])
kmer.length = as.numeric(args[4])
out.path = args[5]
ncore = as.numeric(args[6])

##Example arguments for exploration / debugging
# prot.name = "test_prot"
# out.path = "./test_out"
# prot.path = "./human_proteome_first50.fasta"
# ncore = 1
# kmer.length = 14
# chunk.length = 1e3

###Bring in raw proteome Fasta file
prot.raw = as.data.frame(fread(prot.path,sep='\n',header=F))
start.list = 1:nrow(prot.raw)
start.list = start.list[grepl(">",prot.raw[,1],fixed=T)]

prot.table = as.data.frame(matrix("",length(start.list),2),stringsAsFactors=F)
prot.table[,1] = prot.raw[start.list,]
for(i in 1:(nrow(prot.table) - 1)){
  # print(i)
  prot.table[i,2] = paste(
    prot.raw[
      (start.list[i] + 1):(start.list[i + 1] - 1),1]
    ,collapse="")

}

write.table(prot.table,paste0(out.path,"/",prot.name,"_prot_table.csv"),sep=",",row.names=F)


###Create the KMer set

###Process full length protein sequences
kmer.table = prot.table[
  nchar(prot.table[,2]) >= kmer.length
  & !grepl("[BJOUXZ]",prot.table[,2]),]

# rm(prot.table)
# rm(prot.raw)
# gc()

kmer.set = mclapply(kmer.table[,2],function(x)make.kmer(x,kmer.length),mc.cores = ncore)

kmer.table = kmer.table[rep(c(1:nrow(kmer.table))
    ,times = unlist(lapply(kmer.set,length))),1,drop=F]
colnames(kmer.table)[1] = "FASTA Text"
kmer.table$kmer.seq = unlist(kmer.set)
kmer.table$index = 1:nrow(kmer.table)

save(kmer.table,file = paste0(out.path,"/",prot.name,"_kmerset.Rdata"))



###Create one-hot table representation for all kmers
#Because the one-hot tables are large, and that can cause memory issues for Lua,
#tables are created in chunks of 'chunk.length' rows

peptide.length = kmer.length
for(k in
    1:(ceiling(nrow(kmer.table) / chunk.length))
    ){

  # print(k)
  if(k == ceiling(nrow(kmer.table) / chunk.length)) used.chunk =
      kmer.table[((k - 1) * chunk.length + 1):nrow(kmer.table),]
  if(k < ceiling(nrow(kmer.table) / chunk.length)) used.chunk =
      kmer.table[((k - 1) * chunk.length + 1):(k * chunk.length),]

  used.chunk.oh = blosum.to.roundseq(used.chunk$kmer.seq,sub.mat = onehot.mat)

  used.chunk.oh[,ncol(used.chunk.oh) + 1] = used.chunk$index


  write.table(used.chunk.oh
              ,paste0(out.path,"/",prot.name,"_proteome_onehot_",k,".csv")
              ,sep=",",row.names=F,col.names=F,quote=F)

  rm(used.chunk.oh)
  gc()

}

